/*
 * ACPI based MP arch support.
 */
#include "u.h"
#include "../port/lib.h"
#include "mem.h"
#include "dat.h"
#include "fns.h"
#include "io.h"
#include "mp.h"
#include "acpi.h"

#define dprint	if(debug)print

static int debug = 1;

static int aidentify(void);
static void ainit(void);

PCArch archacpi = {
.id=		"acpi",	
.ident=		aidentify,
.reset=		mpshutdown,
.intrinit=	ainit,
.intrenable=	mpintrenable,
.intron=	lapicintron,
.introff=	lapicintroff,
.fastclock=	i8253read,
.timerset=	lapictimerset,
};

static int
aidentify(void)
{
	char *p;

	/*
	 * Too many options.
	 */
	if((p = getconf("*noacpi")) != nil && strtol(p, 0, 0) != 0)
		return 1;
	if((p = getconf("*nomp")) != nil && strtol(p, 0, 0) != 0)
		return 1;
	if((p = getconf("*9am")) == nil || strtol(p, 0, 0) == 0)
		return 1;

	if(cpuserver && m->havetsc)
		archacpi.fastclock = tscticks;
	/*
	 * We don't know if ACPI will be ok before having
	 * user processes. So say yes; time will tell.
	 * Could check out here if we have rsd ptr.
	 * But that belongs to acpi.c
	 */
	return 0;
}

static void
ainit(void)
{
	Apic *bpapic;

	i8259init();
	syncclock();

	bpapic = mplapics;
	if(bpapic != nil)
		mpstart(bpapic);
	else
		dprint("archacpi: no lapics\n");
}
