/*
 * ACPI 4.0 Support.
 * Still WIP.
 *
 */
#include "kernel.h"
#include "acpi.h"
#include "aml.h"

static int acpidevconf(Aname *dn, int kernconf);

static char*	acpifname = "/dev/acpictl";
static char*	regfname = "/dev/acpiregio";

static Atable*	tfirst;		/* loaded DSDT/SSDT/... tables */
static Atable*	tlast;		/* pointer to last table */
static int	regfd = -1;
static Aobj*	lastreg;	/* last region used for I/O */
static int	lastaccsz;	/* last access size used for I/O */
static Srat*	srat;		/* System resource affinity */
static Msct*	msct;		/* Maximum system characteristics table */
static Gpe	gpes[64];	/* general purpose event handlers */

int acpifd = -1;
int acpisz = 4;			/* sizeof(acpiint) */
int acpiflags;			/* for debug */
char *acpistackbottom;		/* for debug */

void
acpiapics(void)
{
	Aobj *o, *args[Nmargs];
	Aname *n;

	n = amlwalk("/PIC", Walkdev, acpins);
	if(n != nil && n->o != nil && n->o->type == Omethod){
		args[0] = amlnewi(Papic);
		args[1] = nil;
		o = amlcall(n->o, args, nil, acpiflags);
		if(o == nil)
			xprint("call to _PIC failed\n");
		amlclose(o);
	}

}

void
acpictl(char *fmt, ...)
{
	char *msg;
	va_list v;
	int l;

	va_start(v, fmt);
	msg = vsmprint(fmt, v);
	va_end(v);
	if(msg == nil)
		return;
	FLUSH;
	if((acpiflags&Edryrun) == 0){
		l = strlen(msg);
		if(write(acpifd, msg, l) != l)
			xprint("#9/acpictl: '%s': %r\n", msg);
	}else
		dprint("ctl: %s\n", msg);
	free(msg);
}

void
acpiglock(void)
{
	/* get global lock */
}

void
acpigunlock(void)
{
	/* clear global lock */
}

static u64int
l64get(u8int* p)
{
	/*
	 * Doing this as a define
	 * #define l64get(p)	(((u64int)l32get(p+4)<<32)|l32get(p))
	 * causes 8c to abort with "out of fixed registers".
	 */
	return (((u64int)l32get(p+4)<<32)|l32get(p));
}

void
acpigpes(void)
{
	char name[40];
	Aname *n;
	int i, some;

	some = 0;
	for(i = 0; i < nelem(gpes); i++){
		seprint(name, name+sizeof(name), "/GPE/L%02X", i);
		n = amlwalk(name, 0, nil);
		if(n == nil || n->o == nil){
			seprint(name, name+sizeof(name), "/GPE/E%02X", i);
			n = amlwalk(name, 0, nil);
		}
		if(n != nil && n->o != nil){
			some++;
			incref(n->o);
			gpes[i].ho = n->o;
			acpictl("gpe %d %s", i, name);
		}
	}
	if(some == 0)
		dprint("no gpe handlers\n");
}

static void
acpipwr(Aobj *o, int i)
{
	Aobj *v1, *v2;

	v1 = nth(o, 0);
	v2 = nth(o, 1);
	if(v1 != nil && v2 != nil && v1->type == Oint && v2->type == Oint)
		acpictl("pwr %d %#ullx %#ullx", i, v1->ival, v2->ival);
	else
		dprint("pwrs: not a _Sx value: %O\n", o);
}

void
acpipwrs(void)
{
	char name[40];
	Aname *n;
	Aobj *o;
	int i, some;

	some = 0;
	for(i = 0; i <= 5; i++){
		seprint(name, name+sizeof(name), "/_S%d_", i);
		n = amlwalk(name, 0, nil);
		if(n != nil && n->o != nil){
			if(wasamlerror())
				continue;
			o = amlcast(n->o, Opkg);
			popamlerror();
			some++;
			acpipwr(o, i);
			amlclose(o);
		}
	}
	if(some == 0)
		dxprint("no power state values\n");
}

/*
 * Perform I/O within region in access units of accsz bytes.
 * All units in bytes.
 */
long
regio(Aobj *ro, void *p, uintptr off, long len, int accsz, int iswr)
{
	long n;

	if(lastreg != ro || lastaccsz != accsz)
		acpictl("region %s %s %#llux %#ullx %d",
			ro->name->path, acpiregstr(ro->reg.spc),
			ro->reg.base, ro->reg.len, accsz);
	doprint("%s: %s %#p %#lx %#ulx\n", iswr ? "pwrite" : "pread",
		ro->name->path, p, len, off);
	if((acpiflags&Edryrun) == 0){
		if(iswr)
			n = pwrite(regfd, p, len, off);
		else
			n = pread(regfd, p, len, off);
		if(n != len)
			dprint("regio: %r\n");
	}else
		n = len;
	incref(ro);
	amlclose(lastreg);
	lastreg = ro;
	lastaccsz = accsz;
	return n;
}

static Atable*
newtable(uchar *p)
{
	Atable *t;
	Sdthdr *h;

	t = malloc(sizeof(Atable));
	if(t == nil)
		panic("no memory for more aml tables");
	t->tbl = p;
	h = (Sdthdr*)t->tbl;
	t->data = p + Sdthdrsz;
	t->is64 = h->rev >= 2;
	t->dlen = l32get(h->length) - Sdthdrsz;
	memmove(t->sig, h->sig, sizeof(h->sig));
	t->sig[sizeof(t->sig)-1] = 0;
	memmove(t->oemid, h->oemid, sizeof(h->oemid));
	t->oemtblid[sizeof(t->oemtblid)-1] = 0;
	memmove(t->oemtblid, h->oemtblid, sizeof(h->oemtblid));
	t->oemtblid[sizeof(t->oemtblid)-1] = 0;
	t->next = nil;
	if(tfirst == nil)
		tfirst = tlast = t;
	else{
		tlast->next = t;
		tlast = t;
	}
	return t;
}

static Res*
psmallres(uchar *ps, int len)
{
	Res *r;
	uint mask;
	int nb, n, rlen, type;

	if(len < 3)
		return nil;
	type = ps[0]>>3 & 0xF;
	if(0)dxprint("psmallres: type %#x\n", type);
	/*
	 * some devices include many copies of these resources.
	 */
	if(type != ARirq && type != ARdma && type != ARio && type != ARfixio)
		return nil;
	r = amlmalloc(sizeof(Res));
	switch(type){
	case ARirq:
		r->type = Rirq;
		mask = l16get(ps+1);
		memset(r->irq.irqs, -1, sizeof(r->irq.irqs));
		for(nb = n = 0; nb < 16 && mask != 0; nb++){
			if((mask & 1) != 0)
				r->irq.irqs[n++] = nb;
			mask >>= 1;
		}
		if(n == 0){
			free(r);
			return nil;
		}
		if(len < 4)
			r->irq.flags = IntEDGE|IntHIGH;
		else{
			if((ps[3]&8) != 0)
				r->irq.flags = IntLOW;
			else
				r->irq.flags = IntHIGH;
			if((ps[3]&1) != 0)
				r->irq.flags = IntEDGE;
			else
				r->irq.flags = IntLEVEL;
		}
		return r;
	case ARdma:
		r->type = Rdma;
		r->dma.chans = ps[1];
		r->dma.flags = ps[2];
		break;
	case ARio:
		/* should be 8, but some have 4 bytes and a valid ioaddr */
		if(len < 4)
			break;
		r->type = Rio;
		r->io.min = l16get(ps+2);
		r->io.max = r->io.min;
		r->io.align = 1;
		if(len >= 8){
			r->io.max = l16get(ps+4);
			r->io.align = ps[6];
			rlen = ps[7];
			if(r->io.max - r->io.min + 1 != rlen)
				xprint("io: max %#llux min %#llux != len %ux\n",
					r->io.max, r->io.min, rlen);
		}
		return r;
	case ARfixio:
		if(len < 4)
			break;
		r->type = Rio;
		r->io.min = l16get(ps+1);
		rlen = ps[3];
		r->io.max = r->io.min + (rlen - 1);
		r->io.align = 1;
		return r;
	default:
		panic("psmallres bug");
	}
	xprint("psmallres: type %#x len %d failed\n", type, len);
	free(r);
	return nil;
}

static Res*
pasres(Res *r, uchar *ps, int len)
{
	uchar *pe;

	if(ps[0] == ARas16 && len < 16)
		goto Fail;
	if(ps[0] == ARas32 && len < 26)
		goto Fail;
	if(ps[0] == ARas64 && len < 46)
		goto Fail;
	if(ps[0] == ARxas && len != 56)
		goto Fail;
	r->type = ps[3] + Rmemas;
	if(r->type != Rmemas && r->type != Rioas && r->type != Rbusas)
		goto Fail;
	r->as.flags = (ps[4]&7) << 1;
	r->as.flags |= ps[5] << 8;	/* type specific flags */
	pe = ps + len;
	switch(ps[0]){
	case ARas16:
		r->as.mask = l32get(ps+6);
		r->as.min = l32get(ps+8);
		r->as.max = l32get(ps+10);
		r->as.off = l32get(ps+12);
		r->as.len = l32get(ps+14);
		ps += 16;
		break;
	case ARas32:
		r->as.mask = l32get(ps+6);
		r->as.min = l32get(ps+10);
		r->as.max = l32get(ps+14);
		r->as.off = l32get(ps+18);
		r->as.len = l32get(ps+22);
		ps += 26;
		break;
	case ARas64:
		r->as.mask = l64get(ps+6);
		r->as.min = l64get(ps+14);
		r->as.max = l64get(ps+22);
		r->as.off = l64get(ps+30);
		r->as.len = l64get(ps+38);
		ps += 46;
		break;
	case ARxas:
		r->as.rev = ps[6];
		r->as.mask = l64get(ps+8);
		r->as.min = l64get(ps+16);
		r->as.max = l64get(ps+24);
		r->as.off = l64get(ps+32);
		r->as.len = l64get(ps+40);
		r->as.attr = l64get(ps+48);
		return r;
	default:
		goto Fail;
	}
	if(ps[4] & 2)
		r->as.off *= -1;	/* substr. decode */

	if(pe-ps > 2){
		r->as.idx = *ps++;
		r->as.src = amlmalloc(pe-ps+1);
		memmove(r->as.src, ps, pe-ps);
		r->as.src[pe-ps] = 0;
	}
	return r;
Fail:
	xprint("pasres: type %#x len %d failed\n", ps[0], len);
	free(r);
	return nil;
}

static Res*
plargeres(uchar *ps, int len)
{
	Res *r;
	int n;
	int i;
	uchar *pe;
	u64int rlen;

	ps[0] &= 0x7F;
	if(0)dxprint("plargeres: type %#x\n", ps[0]);
	if(len < 3 || ps[0] == 0 || ps[0] == 3 || ps[0] == 4 || ps[0] > 0xB)
		return nil;
	r = amlmalloc(sizeof(Res));
	switch(ps[0]){
	case ARas16:
	case ARas32:
	case ARas64:
	case ARxas:
		return pasres(r, ps, len);
	case ARmem24:
		if(len < 12)
			goto Fail;
		r->type = Rmem;
		r->mem.isrw = ps[3] & 1;
		r->mem.min = l16get(ps+4) << 8;
		r->mem.max = l16get(ps+6) << 8;
		r->mem.align = l16get(ps+8);
		if(r->mem.align == 0)
			r->mem.align = 64 * 1024;
		rlen = l16get(ps+10) << 8;
		if(r->mem.max - r->mem.min != rlen - 1)
			xprint("mem max %llux min %llux != len %llux\n",
				r->mem.max, r->mem.min, rlen);
		break;
	case ARmem32:
		if(len < 20)
			goto Fail;
		r->type = Rmem;
		r->mem.isrw = ps[3] & 1;
		r->mem.min = l32get(ps+4);
		r->mem.max = l32get(ps+8);
		r->mem.align = l32get(ps+12);
		rlen = l32get(ps+16);
		if(r->mem.max - r->mem.min != rlen - 1)
			xprint("mem max %llux min %llux != len %llux\n",
				r->mem.max, r->mem.min, rlen);
		break;
	case ARfixmem32:
		if(len < 12)
			goto Fail;
		r->type = Rmem;
		r->mem.isrw = ps[3] & 1;
		r->mem.min = l32get(ps+4);
		rlen = l32get(ps+8);
		r->mem.max = r->mem.min + (rlen - 1);
		r->mem.align = 1;
		break;
	case ARxirq:
		pe = ps + len;
		if(len < 6)
			goto Fail;
		r->type = Rirq;
		if((ps[3] & 4) != 0)
			r->irq.flags |= IntLOW;
		else
			r->irq.flags |= IntHIGH;
		if((ps[3] & 2) != 0)
			r->irq.flags |= IntEDGE;
		else
			r->irq.flags |= IntLEVEL;
		/*
		if((ps[3] & 8) != 0)
			r->irq.flags |= Ishared;
		if((ps[3] & 1) != 0)
			r->irq.flags |= Iconsume;
		*/
		n = ps[4];
		if(len < 5 + 4*n)
			goto Fail;
		ps += 5;
		for(i = 0; i < n && i < nelem(r->irq.irqs); i++){
			r->irq.irqs[i] = l32get(ps);
			ps += 4;
		}
		if(pe-ps > 2){
			r->irq.idx = *ps++;
			r->irq.src = amlmalloc(pe-ps+1);
			memmove(r->irq.src, ps, pe-ps);
			r->irq.src[pe-ps] = 0;
		}
		break;
	case ARreg:		/* register field */
		if(len != 15)
			goto Fail;
		r->type = Rfield;
		r->field.spc = ps[3];
		r->field.len = ps[4];
		r->field.off = ps[5];
		if(ps[6] != 0)
			r->field.accsz = 1 << (ps[6] + 2);
		r->field.addr = l64get(ps+7);
		break;
	default:
		goto Fail;
	}
	return r;
Fail:
	xprint("plargeres: type %#x len %d failed\n", ps[0], len);
	free(r);
	return nil;
}

/*
 * See if we have a dup resource.
 * We mallocz all resource, it's same to memcmp
 * but for the next pointer.
 */
static int
isdup(Res *l, Res *r)
{
	Res *saved;
	int rc;

	for(; l != nil; l = l->next){
		if(l == r)
			continue;
		saved = l->next;
		l->next = nil;
		rc = memcmp(l, r, sizeof(Res));
		l->next = saved;
		if(rc == 0)
			return 1;
	}
	return 0;
}

static Res*
pres(Res **dr, uchar *ps, uchar *pe)
{
	Res *r;
	Res **rl;
	int len;

	r = nil;
	for(rl = dr; *rl != nil; rl = &(*rl)->next)
		;
	for(; ps < pe; ps += len){
		/* len does not count itself; fix before calling */
		if((ps[0] & 0x80) == 0){
			len = 1 + (ps[0] & 3);
			if(len > pe - ps)
				len = pe - ps;
			*rl = psmallres(ps, len);
		}else if(pe - ps > 3){
			len = 3 + l16get(ps+1);
			if(len > pe - ps)
				len = pe - ps;
			*rl = plargeres(ps, len);
		}else
			break;
		if(*rl != nil)
			if(isdup(*dr, *rl)){
				free(*rl);
				*rl = nil;
			}else
				rl = &(*rl)->next;
	}
	return r;
}

char*
acpiregstr(int id)
{
	static char* dfl[] = {
		"mem", "io", "pcicfg", "embed",
		"smb", "cmos", "pcibar",
	};
	static char buf[20];	/* BUG */

	if(id >= 0 && id < nelem(dfl))
		return dfl[id];
	seprint(buf, buf+sizeof(buf), "spc:%#x", id);
	return buf;
}

static int
Gfmt(Fmt* f)
{
	static char* rnames[] = {
			"mem", "io", "pcicfg", "embed",
			"smb", "cmos", "pcibar", "ipmi"};
	Gas *g;

	g = va_arg(f->args, Gas*);
	switch(g->spc){
	case Rsysmem:
	case Rsysio:
	case Rembed:
	case Rsmbus:
	case Rcmos:
	case Rpcibar:
	case Ripmi:
		fmtprint(f, "[%s ", rnames[g->spc]);
		break;
	case Rpcicfg:
		fmtprint(f, "[pci ");
		fmtprint(f, "dev %#ulx ", (ulong)(g->addr >> 32) & 0xFFFF);
		fmtprint(f, "fn %#ulx ", (ulong)(g->addr & 0xFFFF0000) >> 16);
		fmtprint(f, "adr %#ulx ", (ulong)(g->addr &0xFFFF));
		break;
	case Rfixedhw:
		fmtprint(f, "[hw ");
		break;
	default:
		fmtprint(f, "[spc=%#ux ", g->spc);
	}
	return fmtprint(f, "off %d len %d addr %#ullx sz %d]",
		g->off, g->len, g->addr, g->accsz);
}

/*
 * Apic interrupt flags
 */
static int
Ifmt(Fmt *f)
{
	int r;
	char flg[] = "--";
	r = va_arg(f->args, int);
	if(r == -1)
		return fmtstrcpy(f, "none");
	if((r&IntELMASK) == IntEDGE)
		flg[0] = 'e';
	if((r&IntELMASK) == IntLEVEL)
		flg[0] = 'l';
	if((r&IntPOMASK) == IntHIGH)
		flg[1] = 'h';
	if((r&IntPOMASK) == IntLOW)
		flg[1] = 'l';
	return fmtstrcpy(f, flg);
}

/*
 * print a resource in the format wanted by the kernel.
 */
static int
Rfmt(Fmt *f)
{
	Res *r;
	int i;

	char *nm[] = { "mem", "io", "bus"};

	r = va_arg(f->args, Res*);
	if(r == nil)
		return fmtprint(f, "<null>");
	switch(r->type){
	case Rirq:
		fmtprint(f, "irq %I", r->irq.flags);
		for(i = 0; i < nelem(r->irq.irqs); i++)
			if(r->irq.irqs[i] == ~0)
				break;
			else
				fmtprint(f, " %d", r->irq.irqs[i]);
		break;
	case Rdma:
		return fmtprint(f, "dma %#ux %#ux", r->dma.chans, r->dma.flags);
	case Rio:
		return fmtprint(f, "io %#08llux %#08llux %#ux",
			r->io.min, r->io.max, r->io.align);
	case Rmem:
		return fmtprint(f, "mem %s %#08llux %#08llux %#ux",
			r->mem.isrw ? "rw" : "ro", r->mem.min, r->mem.max,
			r->mem.align);
	case Rioas:
	case Rmemas:
	case Rbusas:
		fmtprint(f, "as %s %#ux", nm[r->type - Rmemas], r->as.flags);
		fmtprint(f, " %#ullx", r->as.mask);
		fmtprint(f, " %#08llux %#08llux", r->as.min, r->as.max);
		fmtprint(f, " %#ullx %#llx", r->as.len, r->as.off);
		fmtprint(f, " %#ullx", r->as.attr);
		break;
	case Rpir:
		fmtprint(f, "pir %2d %d",r->pir.dno, r->pir.pin);
		if(r->pir.src != nil && r->pir.src->type == Odev)
			fmtprint(f, " %ulld -1", r->pir.src->dev.uid);
		else
			fmtprint(f, " -1 %d", r->pir.srcint);
		break;
	case Rbbn:
		return fmtprint(f, "bbn %d", r->bbn);
	default:
		panic("Rfmt with bad resource type");
	}
	return 0;
}

/*
 * Like Rfmt, but for our debug messages and not for
 * the kernel.
 */
static int
Dfmt(Fmt *f)
{
	Res *r;
	int i;
	char *nm[] = { "mem", "io", "bus"};

	r = va_arg(f->args, Res*);
	if(r == nil)
		return fmtprint(f, "<null>");
	switch(r->type){
	case Rirq:
		fmtstrcpy(f, "irqs:");
		for(i = 0; i < nelem(r->irq.irqs); i++)
			if(r->irq.irqs[i] == ~0)
				break;
			else
				fmtprint(f, " %d", r->irq.irqs[i]);
		if(r->irq.src != nil)
			fmtprint(f, " %s[%d]", r->irq.src, r->irq.idx);
		fmtprint(f, " %I", r->irq.flags);
		break;
	case Rdma:
		return fmtprint(f, "dma chmask %#ux flags %#ux",
			r->dma.chans, r->dma.flags);
	case Rio:
		return fmtprint(f, "io [%#08llux-%#08llux]  align %#ux",
			r->io.min, r->io.max, r->io.align);
	case Rmem:
		return fmtprint(f, "mem %s [%#08llux-%#08llux] align %#ux",
			r->mem.isrw ? "rw" : "ro", r->mem.min, r->mem.max,
			r->mem.align);
	case Rioas:
	case Rmemas:
	case Rbusas:
		fmtprint(f, "spc:%s %#ullx@", nm[r->type - Rmemas], r->as.len);
		if(r->as.flags&Asminfixed)
			fmtstrcpy(f, "[");
		else
			fmtstrcpy(f, "(");
		fmtprint(f, "%#08llux-%#08llux", r->as.min, r->as.max);
		if(r->as.flags&Asmaxfixed)
			fmtstrcpy(f, "]");
		else
			fmtstrcpy(f, ")");
		fmtprint(f, " %#+llx flg %#ux", r->as.off, r->as.flags);
		if(r->as.mask)
			fmtprint(f, " mask %#ullx", r->as.mask);
		if(r->as.attr&Muc)
			fmtprint(f, " uc");
		if(r->as.attr&Mwc)
			fmtprint(f, " wc");
		if(r->as.attr&Mwt)
			fmtprint(f, " wt");
		if(r->as.attr&Mwb)
			fmtprint(f, " wb");
		if(r->as.attr&Muce)
			fmtprint(f, " sem");
		if(r->as.attr&Mnvr)
			fmtprint(f, " nvr");
		if(r->as.src != nil)
			fmtprint(f, " %s[%d]", r->as.src, r->as.idx);
		break;
	case Rfield:
		return fmtprint(f, "field %G", &r->field);
	case Rpir:
		fmtprint(f, "pir dev %2d int%c",r->pir.dno, 'a'+r->pir.pin-1);
		if(r->pir.src != nil && r->pir.src->name != nil)
			fmtprint(f, " src %s crs[%d]",
				r->pir.src->name->path, r->pir.idx);
		if(r->pir.srcint >= 0)
			fmtprint(f, " srcint %d", r->pir.srcint);
		break;
	case Rbbn:
		return fmtprint(f, "bbn %d", r->bbn);
	default:
		return fmtprint(f, "<unknown res %#d>", r->type);
	}
	return 0;
}

static void
confaddr(Aobj *o, Aname *n)
{
	Aobj *vo, *r, *v;

	r = nil;
	v = n->o;
	if(v->type == Omethod && v->method.nargs == 0)
		v = r = amlcall(v, nil, nil, acpiflags);
	vo = amlcast(v, Oint);
	o->dev.addr = vo->ival;
	amlclose(vo);
	amlclose(r);
}

static void
confuid(Aobj *o, Aname *n)
{
	Aobj *vo, *r, *v;

	r = nil;
	v = n->o;
	if(v->type == Omethod && v->method.nargs == 0)
		v = r = amlcall(v, nil, nil, acpiflags);
	vo = amlcast(v, Oint);
	o->dev.uid = vo->ival;
	amlclose(vo);
	amlclose(r);
}

static void
confid(Aobj *o, Aname *n)
{
	char id[8];
	Aobj *r, *v;
	static char hex[] = "0123456789ABCDEF";

	if(o->dev.pnpid != nil)
		return;
	r = nil;
	v = n->o;
	if(v->type == Omethod && v->method.nargs == 0)
		v = r = amlcall(v, nil, nil, acpiflags);
	switch(v->type){
	case Ostr:
	case Odecstr:
		kstrdup(&o->dev.pnpid, v->sval);
		break;
	case Oint:
		/* curses taken from bsd code */
		id[0] = '@' + (v->ival >> 2 & 0x1f);
		id[1] = '@' + (v->ival << 3 & 0x18) + (v->ival >> 13 & 0x7);
		id[2] = '@' + (v->ival >> 8 & 0x1f);
		id[3] = hex[v->ival >> 20 & 0xf];
		id[4] = hex[v->ival >> 16 & 0xf];
		id[5] = hex[v->ival >> 28 & 0xf];
		id[6] = hex[v->ival >> 24 & 0xf];
		id[7] = 0;
		kstrdup(&o->dev.pnpid,id);
		break;
	}
	if(r != nil)
		amlclose(r);
}

static void
confname(Aobj *o, Aname *n)
{
	Aobj *vo, *v;

	v = n->o;
	if(o->dev.name == nil){
		vo = amlcast(v, Ostr);
		kstrdup(&o->dev.name,vo->sval);
	}
}

static void
confres(Aobj *o, Aname *n)
{
	Aobj *vo, *ro, *v;
	Res **rl;

	v = n->o;
	ro = nil;
	if(v->type == Omethod && v->method.nargs == 0)
		v = ro = amlcall(v, nil, nil, acpiflags);
	vo = amlcast(v, Obuf);
	if(vo != nil){
		if(strcmp(n->s, "_PRS") == 0)
			rl = &o->dev.other;
		else
			rl = &o->dev.res;
		pres(rl, vo->buf.p, vo->buf.p + vo->buf.len);
		amlclose(vo);
	}
	amlclose(ro);
}

static void
confbbn(Aobj *o, Aname *n)
{
	Aobj *v, *ro, *vo;
	Res *r;

	v = n->o;
	ro = nil;
	if(v->type == Omethod && v->method.nargs == 0)
		v = ro = amlcall(v, nil, nil, acpiflags);
	vo = amlcast(v, Oint);
	if(vo != nil){
		r = amlmalloc(sizeof(Res));
		r->type = Rbbn;
		r->bbn = vo->ival;
		r->next = o->dev.res;
		o->dev.res = r;
		amlclose(vo);
	}
	amlclose(ro);
}

/*
 * We add 1 to the pin number so that pin #1 is inta,
 * matching the numbers used by link devices and we can
 * recognize links w/o uid (uid == 0).
 */
static void
confpir(Aobj *o, Aname *n)
{
	Aobj *ro, *v, *el, *e;
	Res *r, **rl;
	Aname *snam;

	v = n->o;
	ro = nil;
	if(v->type == Omethod && v->method.nargs == 0)
		v = ro = amlcall(v, nil, nil, acpiflags);
	if(v == nil || v->type != Opkg){
		amlclose(ro);
		return;
	}
	for(rl = &o->dev.res; *rl != nil; rl = &(*rl)->next)
		;
	for(el = v->list.hd; el != nil; el = el->next){
		if(el->type != Opkg || el->list.n != 4)
			continue;
		r = *rl = amlmalloc(sizeof(Res));
		rl = &r->next;
		r->type = Rpir;
		e = el->list.hd;
		if(e == nil || e->type != Oint)
			continue;
		/* e->ival is devnb, funnb */
		r->pir.dno = e->ival>>16 & 0xFFFF;
		e = e->next;
		if(e == nil || e->type != Oint)
			continue;
		r->pir.pin = e->ival + 1;
		e = e->next;
		if(e == nil)
			continue;
		r->pir.srcint = -1;
		switch(e->type){
		case Oint:
			r->pir.srcint = 0;
			break;
		case Ostr:
		case Oname:
			snam = amlwalk(e->sval, Walkns, n);
			if(snam == nil)
				xprint("confpir: %s not found\n", e->sval);
			else if(snam->o != nil && snam->o->type == Odev){
				acpidevconf(snam, 0);
				incref(snam->o); /* this breaks stack checks */
				r->pir.src = snam->o;
			}
			else
				xprint("confpir: %s is not a device\n", e->sval);
			break;
		default:
			continue;
		}
		e = e->next;
		if(e == nil || e->type != Oint)
			continue;
		if(r->pir.srcint == 0)
			r->pir.srcint = e->ival;
		else
			r->pir.idx = e->ival;
	}
}

static int
devcall(Aobj *dobj, char *name, Aobj **rp, int type)
{
	Aname *dn, *n;
	Aobj *o;

	dn = dobj->name;
	if(rp != nil)
		*rp = nil;
	n = amlwalk(name, Walkdev, dn);
	if(n == nil)
		return 0;
	if(n->o == nil || n->o->type != Omethod || n->o->method.nargs > 0){
		xprint("acpi: %s: bad %s\n", dn->path, name);
		return -1;
	}
	o = amlcall(n->o, nil, nil, acpiflags);
	if(o == nil){
		xprint("acpi: %s: %s: %s\n", dn->path, name, amlerrstk.msg);
		return -1;
	}
	if(type >= 0 && o->type != type){
		xprint("acpi: %s: %s: not %s\n", dn->path, name, amltname(type));
		amlclose(o);
		return -1;
	}
	if(rp != nil)
		*rp = o;
	else
		amlclose(o);
	return 0;
}

/*
 * XXX: Could be _MAT objects under processors or apcis,
 * that are buffers in MADT format. I've seen none so far.
 *
 * under procs: parse lapic, lsapic, lapicnmi structs.
 * under apic: parse ioapic, iosapic, intsrc and their overrides.
 * in both cases, ignore other entries in the table.
 */
static struct{
	char *s;
	void (*f)(Aobj*, Aname*);
}confs[] = {
	{ "_ADR", confaddr },
	{ "_HID", confid },
	{ "_CID", confid },
	{ "_UID", confuid },
	{ "_STR", confname },
	{ "_DDN", confname },
	{ "_PRS", confres },
	{ "_CRS", confres },
	{ "_PRT", confpir },
	{ "_BBN", confbbn },
};

static void
printres(Res *r, char *p)
{
	for(; r != nil; r = r->next){
		if(1)dprint("res: %s%D\n", p, r);
		switch(r->type){
		case Rirq:
		case Rdma:
		case Rio:
		case Rmem:
		case Rmemas:
		case Rioas:
		case Rbusas:
		case Rpir:
 		case Rbbn:
			acpictl("%s%R\n", p, r);
			break;
		}
	}
}

static int
acpidevconf(Aname *dn, int kconf)
{
	Aname *n;
	Aobj *dobj, *o;
	int rc, i;

	dobj = dn->o;

	if(strncmp(dn->s, "LNK", 3) == 0 && dn->s[3] >= 'A' && dn->s[3] <= 'Z')
		o = nil; /* _STA is meaningless for links */
	else
		if(devcall(dobj, "_STA", &o, Oint) < 0){
			xprint("%s: _STA failed\n", dn->path);
			return -1;
		}
	if(o != nil){
		rc = o->ival;
		amlclose(o);
	}else
		rc = Staok|Stapresent|Staenabled;
	if((acpiflags&Edryrun) == 0 && (rc&Stapresent) == 0){
		dprint("confdev: %s: not present\n", dn->path);
		return -1;
	}
	if(devcall(dobj, "_INI", &o, Oint) < 0){
		xprint("%s: _INI failed\n", dn->path);
		return -1;
	}
	amlclose(o);
	o = nil;
	for(i = 0; i < nelem(confs); i++){
		n = amlwalk(confs[i].s, Walkdev, dn);
		if(n != nil)
			confs[i].f(dn->o, n);
	}
	if(dobj->dev.name == nil)
		kstrdup(&dobj->dev.name, dn->s);	/* get a name */
	if(dobj->dev.pnpid == nil)
		kstrdup(&dn->o->dev.pnpid, "PNPnone");
	if(kconf == 0)
		return 0;
	if((acpiflags&Edryrun) == 0 &&
	   ((rc&Staok) == 0 || (rc&Staenabled) == 0)){
		dprint("device %s is not enabled\n", dn->path);
		return -1;
	}
	acpictl("dev %s %s %s %ulld %#llux\n", dn->path, dobj->dev.name,
		dobj->dev.pnpid, dobj->dev.uid, dobj->dev.addr);
	printres(dobj->dev.res, "");
	printres(dobj->dev.other, "o");
	return 0;
}

/*
 * initialize and configure all devices under n.
 * due to acpi bugs and to our own ones, this may be dangerous.
 */
void
acpiconf(Aname *n)
{
	Aobj *o;
	Aname *cn;
	int isdev;

fprint(2, "acpiconf\n");
	o = n->o;
	isdev = o != nil && o->type == Odev;
	if(isdev){
fprint(2, "acpiconf: dev %s\n", n->s);
		if(acpidevconf(n, 1) < 0)
			return;
		if(n->hasdevs)
			acpictl("[\n");
		dprint("acpiconf: dev %s done\n", n->path);
	}
	for(cn = n->childhd; cn != nil; cn = cn->sibling)
		acpiconf(cn);
	if(isdev && n->hasdevs)
		acpictl("]\n");
}

static void
dumpsrat(Srat *st)
{
	if((acpiflags&Edebugflags) == 0)
		return;
	dprint("acpi: srat:\n");
	for(; st != nil; st = st->next)
		switch(st->type){
		case SRlapic:
			dprint("\tlapic: dom %d apic %d sapic %d clk %d\n",
				st->lapic.dom, st->lapic.apic,
				st->lapic.sapic, st->lapic.clkdom);
			break;
		case SRmem:
			dprint("\tmem: dom %d %#ullx %#ullx %c%c\n",
				st->mem.dom, st->mem.addr, st->mem.len,
				st->mem.hplug?'h':'-',
				st->mem.nvram?'n':'-');
			break;
		case SRlx2apic:
			dprint("\tlx2apic: dom %d apic %d clk %d\n",
				st->lx2apic.dom, st->lx2apic.apic,
				st->lx2apic.clkdom);
			break;
		default:
			dprint("\t<unknown srat entry>\n");
		}
	dprint("\n");
}

int
acpisrat(uchar *p, int len)
{
	Srat **stl, *st;
	uchar *pe;
	int stlen, flags;

	if(srat != nil){
		xprint("acpi: two SRATs?\n");
		return -1;
	}

	stl = &srat;
	pe = p + len;
	for(p += 48; p < pe; p += stlen){
		st = amlmalloc(sizeof(Srat));
		st->type = p[0];
		st->next = nil;
		stlen = p[1];
		switch(st->type){
		case SRlapic:
			st->lapic.dom = p[2] | p[9]<<24| p[10]<<16 | p[11]<<8;
			st->lapic.apic = p[3];
			st->lapic.sapic = p[8];
			st->lapic.clkdom = l32get(p+12);
			if(l32get(p+4) == 0){
				free(st);
				st = nil;
			}
			break;
		case SRmem:
			st->mem.dom = l32get(p+2);
			st->mem.addr = l64get(p+8);
			st->mem.len = l64get(p+16);
			flags = l32get(p+28);
			if((flags&1) == 0){	/* not enabled */
				free(st);
				st = nil;
			}else{
				st->mem.hplug = flags & 2;
				st->mem.nvram = flags & 4;
			}
			break;
		case SRlx2apic:
			st->lx2apic.dom = l32get(p+4);
			st->lx2apic.apic = l32get(p+8);
			st->lx2apic.clkdom = l32get(p+16);
			if(l32get(p+12) == 0){
				free(st);
				st = nil;
			}
			break;
		default:
			xprint("unknown SRAT structure\n");
			free(st);
			st = nil;
		}
		if(st != nil){
			*stl = st;
			stl = &st->next;
		}
	}

	dumpsrat(srat);
	return 0;
}

static void
dumpmsct(Msct *msct)
{
	Mdom *st;

	if((acpiflags&Edebugflags) == 0)
		return;
	dprint("acpi: msct: %d doms %d clkdoms %#ullx maxpa\n",
		msct->ndoms, msct->nclkdoms, msct->maxpa);
	for(st = msct->dom; st != nil; st = st->next)
		dprint("\t[%d:%d] %d maxproc %#ullx maxmmem\n",
			st->start, st->end, st->maxproc, st->maxmem);
	dprint("\n");
}

int
acpimsct(uchar *p, int len)
{
	uchar *pe;
	Mdom **stl, *st;
	int off;

	msct = amlmalloc(sizeof(Msct));
	msct->ndoms = l32get(p+40) + 1;
	msct->nclkdoms = l32get(p+44) + 1;
	msct->maxpa = l64get(p+48);
	msct->dom = nil;
	stl = &msct->dom;
	pe = p + len;
	off = l32get(p+36);
	for(p += off; p < pe; p += 22){
		st = amlmalloc(sizeof(Mdom));
		st->next = nil;
		st->start = l32get(p+2);
		st->end = l32get(p+6);
		st->maxproc = l32get(p+10);
		st->maxmem = l64get(p+14);
		*stl = st;
		stl = &st->next;
	}

	dumpmsct(msct);
	return 0;
}

/*
 * Both DSDT and SSDT
 */
int
acpidsdt(uchar *p, int len)
{
	Atable *t;
	char sig[5];

	if(len < Sdthdrsz)
		return -1;
	memmove(sig, p, 4);
	sig[4] = 0;
	if(strcmp(sig, "DSDT") != 0 && strcmp(sig, "SSDT") != 0)
		return -1;

	t = newtable(p);

	if(wasamlerror())
		return -1;
	/*
	 * Beware that this may be due to a load operation,
	 * which means that evaluation would be nested in
	 * whichever current environment may be found at
	 * amlenv.
	 */

	amleval(t->data, t->data+t->dlen, nil, nil, t, acpiflags);
	popamlerror();
	return 1;	/* keep the aml code allocated */
}

int
acpiinit(void)
{
	fmtinstall('D', Dfmt);
	fmtinstall('R', Rfmt);
	fmtinstall('I', Ifmt);
	fmtinstall('G', Gfmt);
	if((acpiflags&Edryrun) == 0){
		if(acpifd < 0)
			acpifd = open(acpifname, ORDWR);
		if(regfd < 0)
			regfd = open(regfname, ORDWR);
	
		if(regfd < 0 || acpifd < 0){
			xprint("acpi: can't open acpi device: %r\n");
			return -1;
		}
	}
	return 0;
}

