/*
 * ACPI 4.0 constants and definitions.
 */

typedef struct Daddr Daddr;
typedef struct Atable Atable;
typedef struct Facs Facs;
typedef struct Fadt Fadt;
typedef struct Gas Gas;
typedef struct Gpe Gpe;
typedef struct Res Res;
typedef struct Rsdp Rsdp;
typedef struct Sdthdr Sdthdr;
typedef struct Xsdt Xsdt;
typedef struct Aname Aname;
typedef struct Aobj Aobj;
typedef struct Msct Msct;
typedef struct Srat Srat;
typedef struct Mdom Mdom;

enum
{

	Sdthdrsz	= 36,	/* size of SDT header */

	/* ACPI regions. Gas ids */
	Rsysmem	= 0,
	Rsysio,
	Rpcicfg,
	Rembed,
	Rsmbus,
	Rcmos,
	Rpcibar,
	Ripmi,
	Rfixedhw	= 0x7f,

	/* Global lock bits */
	GLpending = 0,
	GLowned	= 1,

	/* resource types */
	Rirq	= 0,
	Rdma,
	Rio,
	Rmem,
	Rmemas,			/* mem address space */
	Rioas,			/* io address space */
	Rbusas,			/* bus address space */
	Rfield,			/* register field */
	/* fake ones, sw */
	Rpir,			/* pci interrupt routing, from _PRT */
	Rbbn,			/* base bus number */

	/* ACPI small resource names */
	ARirq	= 4,
	ARdma	= 5,
	ARio	= 8,
	ARfixio	= 9,

	/* ACPI large resource names (0, 4, and >11 are reserved) */
	ARmem24		= 1,
	ARreg		= 2,		/* register, not region */
	ARmem32		= 5,
	ARfixmem32	= 6,
	ARas32		= 7,
	ARas16		= 8,
	ARxirq		= 9,
	ARas64		= 10,
	ARxas		= 11,

	ARendtag	= 0x78,		/* end resource tag */
	ARdontcksum	= 0,		/* don't checksum the resource */

	/* SRAT types */
	SRlapic = 0,	/* Local apic/sapic affinity */
	SRmem,		/* Memory affinity */
	SRlx2apic,	/* x2 apic affinity */

	/* Arg for _PIC */
	Ppic = 0,	/* PIC interrupt model */
	Papic,		/* APIC interrupt model */
	Psapic,		/* SAPIC interrupt model */

	/* MPS INTI flags; used in Apic flags.
	 * Resources, which do not use the same flags in ACPI, are
	 * recorded also using MPS INT flags, for homegeneity
	 * Flags IntNMI and IntMASKED are fake, used to report that
	 * an interrupt should be NMI or masked.
	 */
	IntPOMASK	= 0x03,		/* polarity conforms to spec of bus */
	IntHIGH		= 0x01,		/* active high */
	IntLOW		= 0x03,		/* active low */
	IntELMASK	= 0x0C,		/* trigger mode of APIC input signals */
	IntEDGE		= 0x04,		/* edge-triggered */
	IntLEVEL	= 0x0C,		/* level-triggered */

	/* ACPI dma resource flags */
	Dmachanmask	= 3<<5,
		Dmachancompat	= 0<<5,
		Dmachana	= 1<<5,
		Dmachanb	= 2<<5,
		Dmachanf	= 3<<5,
	Dmabm		= 1<<1,
	Dmaszmask	= 3,
		Dma8		= 0,
		Dma8and16	= 1,
		Dma16		= 2,

	/* address space resource flags */
	/* type specific flags are kept in the lo byte */
	Asmaxfixed	= 1<<(3+8),
	Asminfixed	= 1<<(2+8),
	Asconsume	= 1<<(0+8),	/* consumes or produces this as? */

	/* ACPI ext. as attrs */
	Muc		= 1,
	Mwc		= 2,
	Mwt		= 4,
	Mwb		= 8,
	Muce		= 0x10,		/* uncached exported, semaphores */
	Mnvr		= 0x8000,

	/* PM1 control */
	Pm1SciEn	= 0x1,		/* Generate SCI and not SMI */

	/* _STA result value flags */
	Stapresent	= 0x1,		/* device is present */
	Staenabled	= 0x2,		/* device is enabled */
	Staok		= 0x8,		/* device is ok */


};

/* Generic address structure. 
 */
struct Gas
{
	u8int	spc;	/* address space id */
	u8int	len;	/* register size in bits */
	u8int	off;	/* bit offset */
	u8int	accsz;	/* 1: byte; 2: word; 3: dword; 4: qword */
	u64int	addr;	/* address (or acpi encoded tbdf + reg) */
};

/* Root system description table pointer.
 * Used to locate the root system description table RSDT
 * (or the extended system description table from version 2) XSDT.
 * The XDST contains (after the DST header) a list of pointers to tables:
 *	- FADT	fixed acpi description table.
 *		It points to the DSDT, AML code making the acpi namespace.
 *	- SSDTs	tables with AML code to add to the acpi namespace.
 *	- pointers to other tables for apics, etc.
 */
struct Rsdp
{
	u8int	signature[8];			/* "RSD PTR " */
	u8int	rchecksum;
	u8int	oemid[6];
	u8int	revision;
	u8int	raddr[4];			/* RSDT */
	u8int	length[4];
	u8int	xaddr[8];			/* XSDT */
	u8int	xchecksum;			/* XSDT */
	u8int	_33_[3];			/* reserved */
};

/* Header for ACPI description tables
 */
struct Sdthdr
{
	u8int	sig[4];			/* "FACP" or whatever */
	u8int	length[4];
	u8int	rev;
	u8int	csum;
	u8int	oemid[6];
	u8int	oemtblid[8];
	u8int	oemrev[4];
	u8int	creatorid[4];
	u8int	creatorrev[4];
};

/* Firmware control structure
 */
struct Facs
{
	u32int	hwsig;
	u32int	wakingv;
	u32int	glock;
	u32int	flags;
	u64int	xwakingv;
	u8int	vers;
	u32int	ospmflags;
};

/*
 * software.
 */

/* System resource affinity table
 */
struct Srat
{
	int	type;
	Srat*	next;
	union{
		struct{
			int	dom;	/* proximity domain */
			int	apic;	/* apic id */
			int	sapic;	/* sapic id */
			int	clkdom;	/* clock domain */
		} lapic;
		struct{
			int	dom;	/* proximity domain */
			u64int	addr;	/* base address */
			u64int	len;
			int	hplug;	/* hot pluggable */
			int	nvram;	/* non volatile */	
		} mem;
		struct{
			int	dom;	/* proximity domain */
			int	apic;	/* x2 apic id */
			int	clkdom;	/* clock domain */
		} lx2apic;
	};
};

/* Maximum System Characteristics table
 */
struct Msct
{
	int	ndoms;		/* number of domains */
	int	nclkdoms;	/* number of clock domains */
	u64int	maxpa;		/* max physical address */

	Mdom*	dom;		/* domain information list */
};

struct Mdom
{
	Mdom*	next;
	int	start;		/* start dom id */
	int	end;		/* end dom id */
	int	maxproc;	/* max processor capacity */
	u64int	maxmem;		/* max memory capacity */
};

/* XSDT/RSDT. 4/8 byte addresses starting at p.
 */
struct Xsdt
{
	int	len;
	int	asize;
	u8int*	p;
};

/* Fixed ACPI description table.
 * Describes implementation and hardware registers.
 * PM* blocks are low level functions.
 * GPE* blocks refer to general purpose events.
 * P_* blocks are for processor features.
 * Has address for the DSDT.
 */
struct Fadt
{
	u32int	facs;
	u32int	dsdt;
	/* 1 reserved */
	u8int	pmprofile;
	u16int	sciint;
	u32int	smicmd;
	u8int	acpienable;
	u8int	acpidisable;
	u8int	s4biosreq;
	u8int	pstatecnt;
	u32int	pm1aevtblk;
	u32int	pm1bevtblk;
	u32int	pm1acntblk;
	u32int	pm1bcntblk;
	u32int	pm2cntblk;
	u32int	pmtmrblk;
	u32int	gpe0blk;
	u32int	gpe1blk;
	u8int	pm1evtlen;
	u8int	pm1cntlen;
	u8int	pm2cntlen;
	u8int	pmtmrlen;
	u8int	gpe0blklen;
	u8int	gpe1blklen;
	u8int	gp1base;
	u8int	cstcnt;
	u16int	plvl2lat;
	u16int	plvl3lat;
	u16int	flushsz;
	u16int	flushstride;
	u8int	dutyoff;
	u8int	dutywidth;
	u8int	dayalrm;
	u8int	monalrm;
	u8int	century;
	u16int	iapcbootarch;
	/* 1 reserved */
	u32int	flags;
	Gas	resetreg;
	u8int	resetval;
	/* 3 reserved */
	u64int	xfacs;
	u64int	xdsdt;
	Gas	xpm1aevtblk;
	Gas	xpm1bevtblk;
	Gas	xpm1acntblk;
	Gas	xpm1bcntblk;
	Gas	xpm2cntblk;
	Gas	xpmtmrblk;
	Gas	xgpe0blk;
	Gas	xgpe1blk;
};

/*
 * ACPI table
 */
struct Atable
{
	Atable*	next;		/* next table in list */
	int	is64;		/* uses 64bits */
	char	sig[5];		/* signature */
	char	oemid[7];	/* oem id str. */
	char	oemtblid[9];	/* oem tbl. id str. */
	uchar*  tbl;		/* pointer to table in memory */
	uchar*	data;		/* pointer to aml bytecodes or nil */
	long	dlen;		/* size of data in table, after Stdhdr */
};


/*
 * Our representation of resource settings. See 6.4 of acpi 4.0.
 */
struct Res
{
	Res*	next;		/* in resource list */
	int	type;		/* resource type */
	union{
		struct{
			u32int	irqs[16];
			uint	flags;
			int	idx;	/* index in src */
			char*	src;	/* where resources come from */
		}irq;
		struct{
			uint	chans;
			uint	flags;
		}dma;
		struct{
			u64int	min;	/* address of the range */
			u64int	max;	/* address of the range */
			uint	align;
			int	isrw;	/* memory only, ro or rw? */
		}io, mem;
		struct{
			u64int	mask;	/* which bits are decoded */
			u64int	min;	/* address on the other side */
			u64int	max;
			u64int	len;	/* can be < max-min+1 if not fixed */
			vlong	off;	/* translation adds this */
			int	idx;	/* index in src */
			char*	src;	/* where resources come from */
			int	flags;
			int	rev;	/* acpi revision id or 0 */
			u64int	attr;
		}as;
		Gas field;

		/* non ACPI resources; invented by us to keep them here */
		struct{
			int	dno;	/* pci device number. */
			int	pin;	/* 1 == INTA */
			int	srcint;	/* global intr. nb. or -1 */
			Aobj*	src;	/* source for interrupt */
			int	idx;
		}pir;			/* pci intr. routing, from _PRT */
		int	bbn;		/* base bus number */
	};
};

struct Gpe
{
	int	nb;		/* event number */
	Aobj*	ho;		/* handler object  */
};

#pragma	varargck	type	"G"	Gas*
#pragma	varargck	type	"I"	int
#pragma	varargck	type	"I"	uint
#pragma	varargck	type	"R"	Res*
#pragma	varargck	type	"D"	Res*

/*
	|c/f2p acpi.c  | c/psort
 */
void	acpiapics(void);
void	acpiconf(Aname *n);
void	acpictl(char *fmt, ...);
int	acpidsdt(uchar *p, int len);
void	acpiglock(void);
void	acpigpes(void);
void	acpigunlock(void);
int	acpiinit(void);
int	acpimsct(uchar *p, int len);
void	acpipwrs(void);
char*	acpiregstr(int id);
int	acpisrat(uchar *p, int len);
long	regio(Aobj *ro, void *p, uintptr off, long len, int accsz, int iswr);

extern int acpifd;
extern int acpisz;		/* number of bytes in integers */
extern int acpiflags;		/* debug flags */
extern char *acpistackbottom;	/* debug */