/*
 * ACPI 4.0 constants and definitions used
 * for the AML interpreter, besides those
 * defined in acpi.h
 */

typedef struct Env Env;
typedef struct Error Error;
typedef struct Mframe Mframe;
typedef struct Aname Aname;
typedef struct Nbuf Nbuf;
typedef struct Ndev Ndev;
typedef struct Nevent Nevent;
typedef struct Nfield Nfield;
typedef struct Nlist Nlist;
typedef struct Nmethod Nmethod;
typedef struct Nmutex Nmutex;
typedef struct Nproc Nproc;
typedef struct Npwr Npwr;
typedef struct Nref Nref;
typedef struct Nreg Nreg;
typedef struct Nscope Nscope;
typedef struct Nslice Nslice;
typedef struct Nstr Nstr;
typedef struct Nthermal Nthermal;
typedef struct Aobj Aobj;
typedef struct Aobjhdr Aobjhdr;
typedef struct Op Op;
typedef struct Xop Xop;

enum
{
	/* AML Operation codes */
	OpZero			= 0x00,
	OpOne			= 0x01,
	OpAlias			= 0x06,
	OpName			= 0x08,
	OpByteprefix		= 0x0A,
	OpWordprefix		= 0x0B,
	OpDwordprefix		= 0x0C,
	OpStrprefix		= 0x0D,
	OpQwordprefix		= 0x0E,
	OpScope			= 0x10,
	OpBuffer		= 0x11,
	OpPackage		= 0x12,
	OpVarpackage		= 0x13,
	OpMethod		= 0x14,
	OpDnameprefix		= 0x2E,
	OpMnameprefix		= 0x2F,
	OpExtprefix		= 0x5B,
		OpMutex			= 0x01,
		OpEvent			= 0x02,
		OpCondrefof		= 0x12,
		OpMkfield		= 0x13,
		OpLoadtable		= 0x1F,
		OpLoad			= 0x20,
		OpStall			= 0x21,
		OpSleep			= 0x22,
		OpAcquire		= 0x23,
		OpSignal			= 0x24,
		OpWait			= 0x25,
		OpReset			= 0x26,
		OpRelease		= 0x27,
		OpFrombcd		= 0x28,
		OpTobcd			= 0x29,
		OpUnload		= 0x2A,
		OpRevision		= 0x30,
		OpDebug			= 0x31,
		OpFatal			= 0x32,
		OpTimer			= 0x33,
		OpOpregion		= 0x80,
		OpField			= 0x81,
		OpDevice			= 0x82,
		OpProcessor		= 0x83,
		OpPwrrsrc		= 0x84,
		OpThermalzone		= 0x85,
		OpIndexfield		= 0x86,
		OpBankfield		= 0x87,
		OpDataregion		= 0x88,
	OpRootchar		= 0x5C,
	OpParentprefix		= 0x5E,
	OpNamechar		= 0x5F,
	OpLocal0		= 0x60,
	OpLocal1		= 0x61,
	OpLocal2		= 0x62,
	OpLocal3		= 0x63,
	OpLocal4		= 0x64,
	OpLocal5		= 0x65,
	OpLocal6		= 0x66,
	OpLocal7		= 0x67,
	OpArg0			= 0x68,
	OpArg1			= 0x69,
	OpArg2			= 0x6A,
	OpArg3			= 0x6B,
	OpArg4			= 0x6C,
	OpArg5			= 0x6D,
	OpArg6			= 0x6E,
	OpStore			= 0x70,
	OpRefof			= 0x71,
	OpAdd			= 0x72,
	OpConcat		= 0x73,
	OpSub			= 0x74,
	OpInc			= 0x75,
	OpDec			= 0x76,
	OpMultiply		= 0x77,
	OpDivide			= 0x78,
	OpShl			= 0x79,
	OpShr			= 0x7A,
	OpAnd			= 0x7B,
	OpNand			= 0x7C,
	OpOr			= 0x7D,
	OpNor			= 0x7E,
	OpXor			= 0x7F,
	OpNot			= 0x80,
	OpLeftsetbit		= 0x81,
	OpRightsetbit		= 0x82,
	OpDerefof		= 0x83,
	OpConcatres		= 0x84,
	OpMod			= 0x85,
	OpNotify			= 0x86,
	OpSizeof			= 0x87,
	OpIndex			= 0x88,
	OpMatch			= 0x89,
	OpMkdwordfield		= 0x8A,
	OpMkwordfield		= 0x8B,
	OpMkbytefield		= 0x8C,
	OpMkbitfield		= 0x8D,
	OpObjecttype		= 0x8E,
	OpMkqwordfield		= 0x8F,
	OpLand			= 0x90,
	OpLor			= 0x91,
	OpLnot			= 0x92,
	OpLequal		= 0x93,
	OpLgreater		= 0x94,
	OpLless			= 0x95,
	OpTobuf			= 0x96,
	OpTodecstr		= 0x97,
	OpTohexstr		= 0x98,
	OpToint			= 0x99,
	OpTostr			= 0x9C,
	OpCopyobject		= 0x9D,
	OpMid			= 0x9E,
	OpContinue		= 0x9F,
	OpIf			= 0xA0,
	OpElse			= 0xA1,
	OpWhile			= 0xA2,
	OpNop			= 0xA3,
	OpReturn		= 0xA4,
	OpBreak			= 0xA5,
	OpBreakpoint		= 0xCC,
	OpOnes			= 0xFF,

	/* OpMatch ops */
	Mtrue	= 0,
	Meq,
	Mle,
	Mlt,
	Mge,
	Mgt,

	/* Field access types */
	AFany	= 0,
	AFbyte,
	AFword,
	AFdword,
	AFqword,
	AFbuffer,

	/* Field update type */
	AFpreserve = 0,
	AFwrones,
	AFwrzeros,

	/* Aobject types returned by AobjectType() */
	Tnone	= 0,
	Tint,
	Tstr,
	Tbuf,
	Tpkg,
	Tfield,
	Tdev,
	Tevent,
	Tmethod,
	Tmutex,
	Treg,
	Tpwr,
	Tproc,
	Tthermal,
	Tslice,
	Thandle,
	Tdbg,

	/* Aobject types used by the interpreter.
	 * Keep this order. amltname() depends on it.
	 * Some of them are not real object types and are
	 * used only for AML argument strings, as parsed by pargs.
	 * Pargs also understands these argument chars:
	 *	- 'l' means packagelen 
	 * 	- 'b' means byte, 'w' word, 'd' d-word,
	 * 	- 'f' fieldlist
	 *	- 's' name string
	 *	- 'n' name or arg or local
	 *	- 't' target (name, perhaps created, dst object, nil)
	 *	- 'o' any object, maybe a name.
	 * Upper case args are evaluated by pargs to obtain the typed arg.
	 * Other args are not evalutated before calling the operation.
	 */
	Onone	=  0,
	Ofree,			/* debug */
	Oslice,			/* buffer field or slice */
	Oscope,			/* scope */
	Omethod,		/* method */
	Odev,			/* device */
	Oproc,			/* processor */
	Opwr,			/* pwrrsrc */
	Othermal,		/* thermal zone */
	Oref,			/* ref to object */
	Olocal,			/* ref to local object */
	Oarg,			/* ref to arg object */
	Ohandle,		/* definition block handle */
	Obuf	= 'B',		/* buf */
	Oevent	= 'E',		/* sleep/wakeup */
	Ofield	= 'F',		/* register field */
	Oint	= 'I',		/* integer */
	Oname	= 'N',		/* a name */
	Omutex	= 'M',		/* mutex */
	Opkg	= 'P',		/* package object (array) */
	Oreg	= 'R',		/* data region */
	Ostr	= 'S',		/* string */
	Odecstr	= 'D',		/* like string, but bytes from buffer */
				/* values are printed in decimal */
	Oany	= 'O',		/* any object after resolving names */

	Nmargs	= 7,		/* max nb. of method arguments */
	Nmlcls	= 8,		/* max nb. of method locals.

	/* Env flags */
	Ewastrue	= 0x0001,	/* last if condition was true */
	Enoeval		= 0x0002,	/* parse but do not eval methods */
	Enocall		= 0x0008,	/* names are names and not calls */
	Eidump		= 0x0010,	/* dump instructions */
	Exdump		= 0x0020,	/* dump actions */
	Emdump		= 0x0040,	/* dump method instructions as well */
	Eedump		= 0x0080,	/* dump environment after ops. */
	Ebdump		= 0x0100,	/* dump ns changes */
	Evdump		= 0x0200,	/* dump general debug diagnostics */
	Eodump		= 0x0400,	/* report object i/o */
	Escheck		= 0x1000,	/* check the stack (slows down a lot) */
	Eabort		= 0x2000,	/* abort on errors */
	Edryrun		= 0x4000,	/* do not perform I/O to /dev */
	Edebugflags	= 0x3FF0,	/* is "debug" on in general? */

	/* arbitrary constants */
	Nargs	= 10,		/* max nb. AML op arguments */
	Nerrors = 32,		/* error stack depth */

	/* amlcpy flag */
	Nocast = 0,
	Cast,

	/* pamlblock flag */
	Dontkeep = 0,
	Keep,

	/* amlwalk flag */
	Walkns = 0,		/* acpi search rules */
	Walkcreating,		/* idem, but create if not found */
	Walkdev,			/* search relative name in device */
};

struct Aname
{
	char	s[5];		/* NNNN\0 */
	char*	path;
	Aobj*	o;
	int	hasdevs;
	Aname*	parent;
	Aname*	childhd;		/* first child */
	Aname*	childtl;		/* last child */
	Aname*	sibling;		/* pointer to next sibling */
	Atable*	table;		/* [ds]sdt table id; for unload */
	Aname*	mnext;		/* in list of per-method binds */
};

struct Nstr
{
	char *aml;
	char *str;
};

struct Error
{
	Label	label[Nerrors];
	int	nerr;
	char	msg[128];
};

typedef void (*Opx)(void);


struct Nbuf
{
	uchar*	p;		/* data */
	ulong	len;		/* number of bytes */
};

struct Nlist
{
	Aobj*	hd;		/* first object */
	int	n;		/* number of objects in packages */
};

/*
 * bit field for a region, or banked-register bit field, or
 * (index,data) register bit field.
 */
struct Nfield
{
	char*	fname;		/* field name */
	int	accsz;		/* access sz in bits for source*/
	int	locking;		/* needs locking */
	int	update;		/* how to update */
	uintptr	off;		/* in bits */
	int	len;		/* in bits */
	Aobj*	reg;		/* region the field is for */
	Aobj*	bank;		/* bank name (for banked field) or nil */
	Aobj*	bankval;	/* bank value (for banked field) */
	Aobj*	idx;		/* index reg. (for index field) or nil */
	Aobj*	idxval;		/* value for index */
	Aobj*	data;		/* data (for index field) or nil */
};

/*
 * buffer field. slice of bits in buffer.
 */
struct Nslice
{
	Aobj*	src;		/* source Obuf object */
	uintptr	off;		/* starting bit index */
	uintptr	len;		/* number of bits */
};

struct Nreg
{
	int	spc;		/* io space */
	u64int	base;		/* address, physical */
	uchar*	p;		/* address, kmapped */
	u64int	len;
};

struct Nscope
{
	char*	name;
	uchar*	pe;		/* end of AML for it */
};

/*
 * ACPI device
 */
struct Ndev
{
	char*	name;
	char*	pnpid;		/* pnp id */
	u64int	uid;		/* unique id valid within same pnpid */
	u64int	addr;		/* address */
	Res*	res;		/* current resource settings */
	Res*	other;		/* other possible res. settings */
	uchar*	pe;
	void	(*handler)(Aobj*, int);	/* for events */
	void*	aux;
};

struct Nmethod
{
	char*	name;
	int	nargs;		/* arity */
	int	isexcl;		/* is serialized? */
	int	synclvl;
	int	trace;		/* 1: on; -1: off; 0: as you were */
	uchar*	ps;		/* aml for this method */
	uchar*	pe;		/* end of aml */
	void	(*f)(Aobj*);	/* C implementation for builtins */
};

struct Nthermal
{
	char*	name;
	void	(*handler)(Aobj*, int);	/* for events */
	void*	aux;
};

struct Npwr
{
	char*	name;
	int	syslvl;
	int	rorder;
};

struct Nproc
{
	char*	name;
	int	id;
	uintptr	baddr;
	int	blen;
	void	(*handler)(Aobj*, int);	/* for events */
	void*	aux;
};

struct Nmutex
{
	int	set;
	int	count;
	int	flags;				/* sync level */
};

struct Nevent
{
	int	fire;
};

struct Nref
{
	Aobj*	o;				/* reference to object */
	int	no;				/* number, for debug */
};

struct Aobjhdr
{
	Ref;
	int	type;				/* for this object */
	Aname*	name;				/* ACPI name if bound or nil */
	Aobj*	next;				/* in list */
};

/*
 * ACPI Aobject.
 */
struct Aobj
{
	Aobjhdr;
	union{
		u64int		ival;		/* doubles as handle */
		char*		sval;		/* strings and names */
		Nbuf		buf;		/* byte buffer */
		Nslice		slice;		/* buffer slice (bits) */
		Nreg		reg;		/* memory, i/o, ... region */
		Nfield		field;		/* slice (bits) of object */
		Nref		oref;		/* local, ref, arg */
		Nlist		list;		/* packages */
		Nscope		scope;		/* for parsing */
		Nmethod		method;
		Ndev		dev;
		Npwr		pwr;
		Nproc		proc;
		Nthermal	thermal;
		Nmutex		mutex;
		Nevent		event;
	};
};

/*
 * AML operation
 */
struct Op
{
	Opx	x;		/* implementation */
	char*	name;		/* debug */
	int	flags;		/* call on dry runs? Print { } on dumps? */
	char*	args;		/* arguments, for AML parsing */
	int	ncalls;		/* how many times used */
};

/*
 * To locate extended operations
 */
struct Xop
{
	int	ic;		/* instruction code */
	Op	op;
};

/*
 * Method activation frame.
 */
struct Mframe
{
	Aobj*	args[Nmargs];	/* method arguments */
	int	nargs;
	Aobj*	lcls[Nmlcls];	/* method locals */
	int	nlcls;
	Aobj**	targs;		/* tmp. method args while parsing them */
	Aname*	binds;		/* done by this method call */
};

/*
 * Evaluation environment supplied to all
 * AML operations.
 * This is used both to narrow the program
 * during parsing and to support evaluation scopes.
 * Envs are nested using the C stack during
 * AML parsing and evaluation.
 */

struct Env
{
	char*	name;		/* debug */
	Env*	prev;		/* in env stack */
	int	ic;		/* instruction code */
	uchar*	ps;		/* start of what's left of program (PC) */
	uchar*	pe;		/* end of program */
	int	lvl;		/* env nesting level (debug) */

	Op*	op;		/* current op */
	uchar*	opp0;		/* operation address 0 */
	Nlist	olist;		/* list of obj in env */
	Aobj*	args[Nargs];	/* arguments for op */
	int	nargs;		/* max number of args used */
	uchar*	argpe;		/* pe resulting from arg 'l' */
	Aobj*	res;		/* result from op */
	
	Mframe*	mp;		/* Method frame pointer */
	Aname*	dot;		/* current acpi name */
	Atable*	table;		/* [ds]sdt table id */
	int	flags;
};

/* print objects, tabs to indent, prog counters */
#pragma	varargck type  "O"   Aobj*
#pragma	varargck type  ">"   int
#pragma	varargck type  "!"   void

#define FmtSharp 8

#define	wasamlerror()	setlabel(&amlerrstk.label[amlerrstk.nerr++])
#define l16get(p)	(((p)[1]<<8)|(p)[0])
#define l32get(p)	(((u32int)l16get(p+2)<<16)|l16get(p))
/*
	|c/f2p amlconv.c amlop.c amlrt.c  | c/psort
 */
void	amlbind(char *s, Aobj *o);
Aobj*	amlcall(Aobj *mo, Aobj **args, Atable *table, int flags);
Aobj*	amlcast(Aobj *o, int twanted);
void	amlclean(Aobj *o);
void	amlclose(Aobj *o);
void	amlcpy(Aobj *o, Aobj *so, int castok);
void	amlenvdump(Env *e);
void	amlerror(char* msg, ...);
char*	amleval(uchar *ps, uchar *pe, Aobj **po, Aname *n, Atable* table, int flags);
void	amlinit(void);
void*	amlmalloc(int l);
Aobj*	amlnew(int t);
Aobj*	amlnewi(int v);
Aobj*	amlnews(char* v);
int	amlnext(void);
Env*	amlpop(void);
void	amlprint(char d, char *fmt, ...);
void	amlpush(Env *new, char* name, uchar *ps, uchar *pe);
Aobj*	amlread(Aobj *o);
void	amlstackcheck(void);
void	amlstackdump(void);
void	amltcheck(Aobj *o, int t, int nullrefok);
char*	amltname(int t);
void	amlunbind(Aname *n);
void	amlunbindtable(Aname *n, Atable *t);
Aname*	amlwalk(char *s, int how, Aname *dot);
void	amlwarn(char* msg, ...);
void	dumpns(Aname *ns, int d, int objs);
void	dumpopstats(void);
void	dumpstats(void);
void	initamlops(void);
int	isamlbranch(void);
Aobj*	opcall(Aobj *o, Aobj **uargs);
Aobj*	pamlblock(int keep);
void	popamlerror(void);
Aobj*	nth(Aobj*, int);

#define dxprint(...)	amlprint('x', __VA_ARGS__)
#define diprint(...)	amlprint('i', __VA_ARGS__)
#define daprint(...)	amlprint('a', __VA_ARGS__)
#define dbprint(...)	amlprint('b', __VA_ARGS__)
#define doprint(...)	amlprint('o', __VA_ARGS__)
#define dprint(...)	amlprint('d', __VA_ARGS__)
#define xprint(...)	amlprint(0, __VA_ARGS__)

extern	Aname*	acpins;
extern	Env*	amlenv;
extern	Error	amlerrstk;
extern	Aobj*	acpilock;

